From 9ea4f3ca0a18bbb22add1acbeba422c3d4f30a9d Mon Sep 17 00:00:00 2001
From: Wayne Xia <xiawenc@linux.vnet.ibm.com>
Date: Tue, 5 Jul 2011 00:22:33 +0800
Subject: [Qemu-devel] [PATCH v5] showing a splash picture when start

Made an option to let qemu pass a picture to bios, let the bios show it as a
logo. By default it is off, enable it as following
    -boot splash=P<,splash-time=T>
P is the file name or an absolute path, specifying it would enable log. It
could be a jpg/bmp file, or a compressed jpg/bmp file with suffix .7z/.7Z
by LZMA algorithm.
T have a max value of 0xffff, unit is ms, and its predefined value is 2500ms.

Signed-off-by: Wayne Xia <xiawenc@linux.vnet.ibm.com>
---
 hw/fw_cfg.c   |  161 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++-
 qemu-config.c |   27 ++++++++++
 sysemu.h      |    3 +
 vl.c          |   17 ++++++-
 4 files changed, 206 insertions(+), 2 deletions(-)

diff --git a/hw/fw_cfg.c b/hw/fw_cfg.c
index 85c8c3c..8b2dcf6 100644
--- a/hw/fw_cfg.c
+++ b/hw/fw_cfg.c
@@ -26,6 +26,7 @@
 #include "isa.h"
 #include "fw_cfg.h"
 #include "sysbus.h"
+#include "qemu-error.h"
 
 /* debug firmware config */
 //#define DEBUG_FW_CFG
@@ -56,6 +57,164 @@ struct FWCfgState {
     Notifier machine_ready;
 };
 
+#define INV_FILE -1
+#define LZMA_FILE 0
+#define JPG_FILE 1
+#define BMP_FILE 2
+
+static FILE *probe_splashfile(char *filename, int *file_sizep, int *file_typep)
+{
+    FILE *fp = NULL;
+    int fop_ret;
+    int file_size;
+    int file_type = INV_FILE;
+    unsigned char buf[4] = {0, 0, 0, 0};
+    unsigned int filehead_value = 0;
+    int bmp_bpp;
+    char *ext_name_p;
+
+    fp = fopen(filename, "rb");
+    if (fp == NULL) {
+        error_report("failed to open file '%s'.", filename);
+        return fp;
+    }
+    /* check file size */
+    fseek(fp, 0L, SEEK_END);
+    file_size = ftell(fp);
+    if (file_size < 2) {
+        error_report("file size is less than 2 bytes '%s'.", filename);
+        fclose(fp);
+        fp = NULL;
+        return fp;
+    }
+    /* check filetype */
+    /* LZMA compressed file have no magic ID, check file name */
+    ext_name_p = strrchr(filename, '.');
+    if (ext_name_p != NULL) {
+        if ((strcmp(ext_name_p, ".7z") == 0)
+                || (strcmp(ext_name_p, ".7Z") == 0)) {
+            file_type = LZMA_FILE;
+        }
+    }
+    /* check magic ID */
+    if (file_type == INV_FILE) {
+        fseek(fp, 0L, SEEK_SET);
+        fop_ret = fread(buf, 1, 2, fp);
+        filehead_value = (buf[0] + (buf[1] << 8)) & 0xffff;
+        if (filehead_value == 0xd8ff) {
+            file_type = JPG_FILE;
+        } else {
+            if (filehead_value == 0x4d42) {
+                file_type = BMP_FILE;
+            }
+        }
+    }
+    if (file_type == INV_FILE) {
+        error_report("'%s' is not jpg/bmp/7z file,head:0x%x.",
+                         filename, filehead_value);
+        fclose(fp);
+        fp = NULL;
+        return fp;
+    }
+
+    /* check BMP bpp */
+    if (file_type == BMP_FILE) {
+        fseek(fp, 28, SEEK_SET);
+        fop_ret = fread(buf, 1, 2, fp);
+        bmp_bpp = (buf[0] + (buf[1] << 8)) & 0xffff;
+        if (bmp_bpp != 24) {
+            error_report("only 24bpp bmp file is supported.");
+            fclose(fp);
+            fp = NULL;
+            return fp;
+        }
+    }
+    /* return values */
+    *file_sizep = file_size;
+    *file_typep = file_type;
+    return fp;
+}
+
+static void fw_cfg_bootsplash(FWCfgState *s)
+{
+    int boot_splash_time = 2500; /* predifined time */
+    const char *boot_splash_filename = NULL; /*default is off */
+    char *p;
+    char *filename;
+    FILE *fp;
+    int fop_ret;
+    int file_size;
+    int file_type = -1;
+    const char *temp;
+
+    /* get user configuration */
+    QemuOptsList *plist = qemu_find_opts("boot-opts");
+    QemuOpts *opts = QTAILQ_FIRST(&plist->head);
+    if (opts != NULL) {
+        temp = qemu_opt_get(opts, "splash");
+        if (temp != NULL) {
+            boot_splash_filename = temp;
+        }
+        temp = qemu_opt_get(opts, "splash-time");
+        if (temp != NULL) {
+            p = (char *)temp;
+            boot_splash_time = strtol(p, (char **)&p, 10);
+        }
+    }
+
+    /* check user configuration */
+    if (boot_splash_filename == NULL) {
+        /* do nothing, directly return */
+        return;
+    }
+    if (boot_splash_time > 0xffff) {
+        error_report("splash time is big than 65535, force it to 65535.");
+        boot_splash_time = 65535;
+    }
+    filename = qemu_find_file(QEMU_FILE_TYPE_BIOS, boot_splash_filename);
+    if (filename == NULL) {
+        error_report("failed to find file '%s'.", boot_splash_filename);
+        return;
+    }
+
+    /* probing the file */
+    fp = probe_splashfile(filename, &file_size, &file_type);
+    if (fp == NULL) {
+        qemu_free(filename);
+        return;
+    }
+
+    /* loading file data */
+    if (boot_splash_filedata != NULL) {
+        qemu_free(boot_splash_filedata);
+    }
+    boot_splash_filedata = qemu_malloc(file_size);
+    boot_splash_filedata_size = file_size;
+    fseek(fp, 0L, SEEK_SET);
+    fop_ret = fread(boot_splash_filedata, 1, file_size, fp);
+    fclose(fp);
+
+    /* insert data */
+    if (file_type == JPG_FILE) {
+        fw_cfg_add_file(s, "bootsplash.jpg",
+                boot_splash_filedata, boot_splash_filedata_size);
+    } else {
+        if (file_type == BMP_FILE) {
+            fw_cfg_add_file(s, "bootsplash.bmp",
+                    boot_splash_filedata, boot_splash_filedata_size);
+        } else {
+            fw_cfg_add_file(s, "bootsplash.lzma",
+                    boot_splash_filedata, boot_splash_filedata_size);
+        }
+    }
+    /* use little endian format */
+    qemu_extra_params_fw[0] = (uint8_t)(boot_splash_time & 0xff);
+    qemu_extra_params_fw[1] = (uint8_t)((boot_splash_time >> 8) & 0xff);
+    fw_cfg_add_file(s, "qemu_extra_params_fw.cfg", qemu_extra_params_fw, 4);
+    qemu_free(filename);
+}
+
+
 static void fw_cfg_write(FWCfgState *s, uint8_t value)
 {
     int arch = !!(s->cur_entry & FW_CFG_ARCH_LOCAL);
@@ -352,7 +511,7 @@ FWCfgState *fw_cfg_init(uint32_t ctl_port, uint32_t data_port,
     fw_cfg_add_i16(s, FW_CFG_NB_CPUS, (uint16_t)smp_cpus);
     fw_cfg_add_i16(s, FW_CFG_MAX_CPUS, (uint16_t)max_cpus);
     fw_cfg_add_i16(s, FW_CFG_BOOT_MENU, (uint16_t)boot_menu);
-
+    fw_cfg_bootsplash(s);
 
     s->machine_ready.notify = fw_cfg_machine_ready;
     qemu_add_machine_init_done_notifier(&s->machine_ready);
diff --git a/qemu-config.c b/qemu-config.c
index c63741c..5b0b4a6 100644
--- a/qemu-config.c
+++ b/qemu-config.c
@@ -469,6 +469,32 @@ static QemuOptsList qemu_machine_opts = {
     },
 };
 
+QemuOptsList qemu_boot_opts = {
+    .name = "boot-opts",
+    .head = QTAILQ_HEAD_INITIALIZER(qemu_boot_opts.head),
+    .desc = {
+        /* the three names below are not used now */
+        {
+            .name = "order",
+            .type = QEMU_OPT_STRING,
+        }, {
+            .name = "once",
+            .type = QEMU_OPT_STRING,
+        }, {
+            .name = "menu",
+            .type = QEMU_OPT_STRING,
+        /* following are really used */
+        }, {
+            .name = "splash",
+            .type = QEMU_OPT_STRING,
+        }, {
+            .name = "splash-time",
+            .type = QEMU_OPT_STRING,
+        },
+        { /*End of list */ }
+    },
+};
+
 static QemuOptsList *vm_config_groups[32] = {
     &qemu_drive_opts,
     &qemu_chardev_opts,
@@ -484,6 +510,7 @@ static QemuOptsList *vm_config_groups[32] = {
 #endif
     &qemu_option_rom_opts,
     &qemu_machine_opts,
+    &qemu_boot_opts,
     NULL,
 };
 
diff --git a/sysemu.h b/sysemu.h
index d3013f5..6817752 100644
--- a/sysemu.h
+++ b/sysemu.h
@@ -123,6 +123,9 @@ extern int no_shutdown;
 extern int semihosting_enabled;
 extern int old_param;
 extern int boot_menu;
+extern uint8_t *boot_splash_filedata;
+extern int boot_splash_filedata_size;
+extern uint8_t qemu_extra_params_fw[4];
 extern QEMUClock *rtc_clock;
 
 #define MAX_NODES 64
diff --git a/vl.c b/vl.c
index 52402a2..fe03c99 100644
--- a/vl.c
+++ b/vl.c
@@ -228,6 +228,9 @@ int ctrl_grab = 0;
 unsigned int nb_prom_envs = 0;
 const char *prom_envs[MAX_PROM_ENVS];
 int boot_menu;
+uint8_t *boot_splash_filedata;
+int boot_splash_filedata_size;
+uint8_t qemu_extra_params_fw[4];
 
 typedef struct FWBootEntry FWBootEntry;
 
@@ -293,6 +296,14 @@ static struct {
     { .driver = "qxl-vga",              .flag = &default_vga       },
 };
 
+static void res_free(void)
+{
+    if (boot_splash_filedata != NULL) {
+        qemu_free(boot_splash_filedata);
+        boot_splash_filedata = NULL;
+    }
+}
+
 static int default_driver_check(QemuOpts *opts, void *opaque)
 {
     const char *driver = qemu_opt_get(opts, "driver");
@@ -2314,7 +2325,8 @@ int main(int argc, char **argv, char **envp)
             case QEMU_OPTION_boot:
                 {
                     static const char * const params[] = {
-                        "order", "once", "menu", NULL
+                        "order", "once", "menu",
+                        "splash", "splash-time", NULL
                     };
                     char buf[sizeof(boot_devices)];
                     char *standard_boot_devices;
@@ -2357,6 +2369,8 @@ int main(int argc, char **argv, char **envp)
                                 exit(1);
                             }
                         }
+                        qemu_opts_parse(qemu_find_opts("boot-opts"),
+                                        optarg, 0);
                     }
                 }
                 break;
@@ -3318,6 +3332,7 @@ int main(int argc, char **argv, char **envp)
     main_loop();
     quit_timers();
     net_cleanup();
+    res_free();
 
     return 0;
 }
-- 
1.7.6.rc1.1.g2c162b
